/*-
 * Copyright (c) 1998 Kazutaka YOKOTA <yokota@zodiac.mech.utsunomiya-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer as
 *    the first lines of this file unmodified.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: $
 */

#include "mouse.h"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/conf.h>
#include <sys/proc.h>
#include <sys/uio.h>

#include <dev/mouse/vmousereg.h>

static int		mice = 0;
static vmouse_t		**mouse;
static struct cdevsw	**moucdevsw;

#define ARRAY_DELTA	4

static void
mou_realloc_array(void)
{
	struct cdevsw **new_cdevsw;
	vmouse_t **new_mouse;
	int newsize;
	int s;

	s = spltty();
	newsize = ((mice + ARRAY_DELTA)/ARRAY_DELTA)*ARRAY_DELTA;
	new_cdevsw = malloc(sizeof(*new_cdevsw)*newsize, M_DEVBUF, M_WAITOK);
	new_mouse = malloc(sizeof(*new_mouse)*newsize, M_DEVBUF, M_WAITOK);
	bzero(new_mouse, sizeof(*new_mouse)*newsize);
	bzero(new_cdevsw, sizeof(*new_cdevsw)*newsize);
	if (mice > 0) {
		bcopy(mouse, new_mouse, sizeof(*mouse)*mice);
		bcopy(moucdevsw, new_cdevsw, sizeof(*moucdevsw)*mice);
		free(mouse, M_DEVBUF);
		free(moucdevsw, M_DEVBUF);
	}
	mouse = new_mouse;
	moucdevsw = new_cdevsw;
	mice = newsize;
	splx(s);

	if (bootverbose)
		printf("mou: new array size %d\n", mice);
}

/*
 * Virtual mouse cdev driver functions
 * The virtual mouse driver dispatches driver functions to
 * appropriate subdrivers.
 */

#define MOU_UNIT(dev)	(minor(dev) >> 1)
#define MOU_MKMINOR(u, b) (((u) << 1) | ((b) ? 0 : 1))

static d_open_t		mouopen;
static d_close_t	mouclose;
static d_read_t		mouread;
static d_write_t	mouwrite;
static d_ioctl_t	mouioctl;
static d_reset_t	moureset;
static d_devtotty_t	moudevtotty;
static d_poll_t		moupoll;
static d_mmap_t		moummap;

#define CDEV_MAJOR	138	/* XXX */

static struct cdevsw mou_cdevsw = {
	mouopen,	mouclose,	mouread,	mouwrite,   /* ??? */
	mouioctl,	nullstop,	moureset,	moudevtotty,
	moupoll,	moummap,	nostrategy,	"mouse",
	NULL,		-1,		nodump,		nopsize,
};

static void
vmouattach(void *arg)
{
	static int mou_devsw_installed = FALSE;
	dev_t dev;

	if (!mou_devsw_installed) {
		dev = makedev(CDEV_MAJOR, 0);
		cdevsw_add(&dev, &mou_cdevsw, NULL);
		mou_devsw_installed = TRUE;
	}
}

PSEUDO_SET(vmouattach, mouse);

/* initialize the vmouse_t structure */
void
mou_init_struct(vmouse_t *mou, char *name, int type, int unit, int config)
{
	mou->ms_flags = 0;
	mou->ms_name = name;		/* driver name */
	mou->ms_type = type;		/* MOUSE_MODEL_XXX */
	mou->ms_unit = unit;		/* unit # */
	mou->ms_config = config;
	mou->ms_data = NULL;
}

int
mou_attach(dev_t dev, vmouse_t *mou, struct cdevsw *cdevsw)
{
	int s;
	int i;

	s = spltty();
	for (i = 0; i < mice; ++i) {
		if (mouse[i] == NULL)
			break;
	}
	if (i >= mice)
		mou_realloc_array();
	/* assert(i < mice) */

	mou->ms_index = i;		/* vmouse index */
	mou->ms_minor = minor(dev);	/* save the subdevice's minor # */
	mouse[i] = mou;
	moucdevsw[i] = cdevsw;
	splx(s);

	printf("mouse%d at %s%d\n", mou->ms_index, mou->ms_name, mou->ms_unit);
	return 0;
}

int
mou_detach(dev_t dev, vmouse_t *mou, struct cdevsw *cdevsw)
{
	int s;

	if (mou->ms_index >= mice)
		return EINVAL;
	if (mouse[mou->ms_index] != mou)
		return EINVAL;
	if (moucdevsw[mou->ms_index] != cdevsw)
		return EINVAL;
	s = spltty();
	mouse[mou->ms_index] = NULL;
	moucdevsw[mou->ms_index] = NULL;
	splx(s);
	return 0;
}

static int
mouopen(dev_t dev, int flag, int mode, struct proc *p)
{
	int unit;

	unit = MOU_UNIT(dev);
	if (unit >= mice)
		return ENXIO;
	if (moucdevsw[unit] == NULL)
		return ENXIO;
	return (*moucdevsw[unit]->d_open)(makedev(0, mouse[unit]->ms_minor),
					  flag, mode, p);
}

static int
mouclose(dev_t dev, int flag, int mode, struct proc *p)
{
	int unit;

	unit = MOU_UNIT(dev);
	if (moucdevsw[unit] == NULL)
		return ENXIO;
	return (*moucdevsw[unit]->d_close)(makedev(0, mouse[unit]->ms_minor),
					   flag, mode, p);
}

static int
mouread(dev_t dev, struct uio *uio, int flag)
{
	int unit;

	unit = MOU_UNIT(dev);
	if (moucdevsw[unit] == NULL)
		return ENXIO;
	return (*moucdevsw[unit]->d_read)(makedev(0, mouse[unit]->ms_minor),
					  uio, flag);
}

static int
mouwrite(dev_t dev, struct uio *uio, int flag)
{
	int unit;

	unit = MOU_UNIT(dev);
	if (moucdevsw[unit] == NULL)
		return ENXIO;
	return (*moucdevsw[unit]->d_write)(makedev(0, mouse[unit]->ms_minor),
					   uio, flag);
}

static int
mouioctl(dev_t dev, u_long cmd, caddr_t arg, int flag, struct proc *p)
{
	int unit;

	unit = MOU_UNIT(dev);
	if (moucdevsw[unit] == NULL)
		return ENXIO;
	return (*moucdevsw[unit]->d_ioctl)(makedev(0, mouse[unit]->ms_minor),
					   cmd, arg, flag, p);
}

static int
moureset(dev_t dev)
{
	int unit;

	unit = MOU_UNIT(dev);
	if (moucdevsw[unit] == NULL)
		return ENXIO;
	return (*moucdevsw[unit]->d_reset)(makedev(0, mouse[unit]->ms_minor));
}

static struct tty
*moudevtotty(dev_t dev)
{
	int unit;

	unit = MOU_UNIT(dev);
	if (moucdevsw[unit] == NULL)
		return NULL;
	return (*moucdevsw[unit]->d_devtotty)(makedev(0, mouse[unit]->ms_minor));
}

static int
moupoll(dev_t dev, int event, struct proc *p)
{
	int unit;

	unit = MOU_UNIT(dev);
	if (moucdevsw[unit] == NULL)
		return ENXIO;
	return (*moucdevsw[unit]->d_poll)(makedev(0, mouse[unit]->ms_minor),
					  event, p);
}

static int
moummap(dev_t dev,  vm_offset_t offset, int nprot)
{
	int unit;

	unit = MOU_UNIT(dev);
	if (moucdevsw[unit] == NULL)
		return ENXIO;
	return (*moucdevsw[unit]->d_mmap)(makedev(0, mouse[unit]->ms_minor),
					  offset, nprot);
}
